/*++
/* NAME
/*      simfluid.c
/* SUMMARY
/*  Simulate throughput of fluid flow line with general up and down times
/*  This version includes quality losses
/* SYNOPSIS
/* DESCRIPTION
/* SEE ALSO
/* AUTHOR(S)
/*      I.Adan
/*      Department of Mathematics and Computer Science
/*      Eindhoven University of technology
/*      Den Dolech 2, P.O. Box 513, 5600 MB Eindhoven, The Netherlands
/* CREATION DATE
/*      15 Nov 2008
/* LAST MODIFICATION
/*      15 Nov 2008
/* VERSION/RELEASE
/*      1.00
/*--*/

#include <stdlib.h>
#include <math.h>

 /* some silly macros */

#define MAXN        100
#define MAXR        1000
#define MAXI        100
#define MAXS        1000
#define MAX(x,y)    (((x) > (y)) ? (x) : (y))
#define MIN(x,y)    (((x) < (y)) ? (x) : (y))
#define SQ(x)       ((x) * (x))
#define DOWN        0
#define UP      1
#define STARVED     2
#define BLOCKED     3
#define EMPTY       0
#define NEUTRAL     1
#define FULL        2
#define MACH        1
#define BUF     2
#define HIS     3
#define INF     10e+10				/* just a very large number */
#define EPS     10e-8				/* just a very small number */

 /* globals */
long    seed;				/* seed of random generator */
double  t,				/* current time */
        tprev,				/* time of previous event */
        totalprod,			/* total production */
        area1[MAXN],			/* area of buffer content process over time */
        area2[MAXN],			/* area of buffer content process over time for second moment */
        vmaxlength[MAXN],		/* duration machine is producing at max rate */
        runlength,			/* runlength of simulation */
        b[MAXN],			/* current buffer level */
        evmach[MAXN],			/* next event time of machine */
        evbuf[MAXN],			/* next event time of buffer */
        evres[MAXN],			/* residual next event time of machine */
        v[MAXN],			/* actual rate of machine */
        vnet[MAXN],			/* net rate at which buffer content changes */
        TH[MAXR],			/* throughput in run */
        L1[MAXR][MAXN],			/* mean buffer contents in run */
        L2[MAXR][MAXN],			/* second moment buffer contents in run */
        LS[MAXR][MAXN],			/* standard deviation buffer contents in run */
        FM[MAXR][MAXN],			/* fraction of time that machine is producing at max rate in run */
        VM[MAXR][MAXN],			/* mean durarion of period during which machine is producing at max rate in run */
        tsample;			/* sampling event for buffer content */
int     vmaxnr[MAXN];			/* number of periods machine is producing at max rate */

 /* model data */
int     N,				/* length of fluid flow line */
        nrrun,				/* number of runs */
        stbuf[MAXN],			/* state of buffer */
        stmach[MAXN],			/* state of machine */
        histo[MAXN][MAXI],		/* histogram of buffer content */
        nrsample;			/* number of samples for buffer content */
double  upmean[MAXN],			/* mean up time (operational dependent) */
        downmean[MAXN],			/* mean down time */
        upsd[MAXN],			/* standard deviation of up time */
        downsd[MAXN],			/* standard deviation of down time */
        upm3[MAXN],			/* 3-rd moment of up time */
        downm3[MAXN],			/* 3-rd moment of down time */
        vmax[MAXN],			/* max rate of machine */
        q[MAXN],			/* quality of machine: fraction of production that meets quality standard */
        K[MAXN];			/* buffer size (> 0) */


 /* declarations */
void    input (),
        output (),
        initialization (),
        update (),
        updatemach (int),
        updatebuf (int),
        updatestat (int),
        printstatus (),
        updatehisto (),
        printhisto ();
int     nextevent (int *);
double  uptime (int),
        downtime (int),
        exponential (double);

main ()
{
    int     j,
            run;

    input ();
    initialization ();

    for (run = 0; run <= nrrun; run++)
      {
	  while (t < runlength)
	    {
		tprev = t;
		if (nextevent (&j) == MACH)
		  {
		      t = evmach[j];		/* advance time */
		      update ();
		      updatemach (j);		/* do book keeping because of machine event */
		  }
		else if (nextevent (&j) == BUF)	/* nextevent(j) == BUF */
		  {
		      t = evbuf[j];		/* advance time */
		      update ();
		      updatebuf (j);		/* do book keeping because of buffer event */
		  }
		else if (nextevent (&j) == HIS)	/* nextevent(j) == HIS */
		  {
		      t = tsample;		/* advance time */
		      update ();
		      updatehisto ();
		  }
		else
		  {
		      printf ("nextevent is not MACH nor BUF: impossible\n");
		      exit (1);
		  }
	    }
	  updatestat (run);
      }
    output ();

    printhisto ();
}

void input ()
{
    int     i,
            j;
    char    str[50];

    scanf ("%s%d", str, &N);

    for (i = 0; i <= N; i++)
      {
	  scanf ("%s%d", str, &j);
	  scanf ("%s%lf", str, &upmean[j]);
	  scanf ("%s%lf", str, &upsd[j]);
	  scanf ("%s%lf", str, &upm3[j]);
	  scanf ("%s%lf", str, &downmean[j]);
	  scanf ("%s%lf", str, &downsd[j]);
	  scanf ("%s%lf", str, &downm3[j]);
	  scanf ("%s%lf", str, &vmax[j]);
	  scanf ("%s%lf", str, &q[j]);
      }

    for (i = 1; i <= N; i++)
      {
	  scanf ("%s%d", str, &j);
	  scanf ("%s%lf", str, &K[j]);
      }

    scanf ("%s%ld", str, &seed);
    scanf ("%s%lf", str, &runlength);
    scanf ("%s%d", str, &nrrun);
}

void output ()
{
    int     i,
            run;
    double  THav,			/* average throughtput */
            Lav,			/* average buffer content */
            Sav;			/* average flow time */
    double  mean,
            square,
            stdev;


    /* first print input parameters */

    printf ("length     %d\n", N);

    for (i = 0; i <= N; i++)
      {
	  printf ("\n");
	  printf ("machine      %d\n", i);
	  printf ("uptimemean   %g\n", upmean[i]);
	  printf ("uptimesd %g\n", upsd[i]);
	  printf ("uptimem3 %g\n", upm3[i]);
	  printf ("downtimemean %g\n", downmean[i]);
	  printf ("downtimesd   %g\n", downsd[i]);
	  printf ("downtimem3 %g\n", downm3[i]);
	  printf ("ratemax      %g\n", vmax[i]);
      }

    printf ("\n");
    for (i = 1; i <= N; i++)
      {
	  printf ("buffer       %d\n", i);
	  printf ("capacity %g\n", K[i]);
      }

    printf ("\n");
    printf ("seed       %ld\n", seed);
    printf ("runlength  %g\n", runlength);
    printf ("nrrun      %d\n", nrrun);

    /* now print performance estimates */

    for (mean = square = 0.0, run = 1; run <= nrrun; run++)
      {
	  mean += TH[run];
	  square += SQ (TH[run]);
      }
    mean /= nrrun;
    square /= nrrun;
    stdev = sqrt (square - SQ (mean));
    THav = mean;
    printf ("\n");
    printf ("throughput         %g      stdev %g\n", mean, stdev);

    for (Lav = 0.0, i = 1; i <= N; i++)
      {
	  for (mean = square = 0.0, run = 1; run <= nrrun; run++)
	    {
		mean += L1[run][i];
		square += SQ (L1[run][i]);
	    }
	  mean /= nrrun;
	  square /= nrrun;
	  stdev = sqrt (square - SQ (mean));
	  Lav += mean;
	  printf ("mean content buffer %d   %g  stdev %g\n", i, mean, stdev);

	  for (mean = square = 0.0, run = 1; run <= nrrun; run++)
	    {
		mean += LS[run][i];
		square += SQ (LS[run][i]);
	    }
	  mean /= nrrun;
	  square /= nrrun;
	  stdev = sqrt (square - SQ (mean));
	  printf ("stdv content buffer %d   %g  stdev %g\n", i, mean, stdev);
      }
    Sav = Lav / THav;				/* Little's law */
    printf ("mean flow time     %g\n", Sav);

    for (i = 0; i <= N; i++)
      {
	  for (mean = square = 0.0, run = 1; run <= nrrun; run++)
	    {
		mean += FM[run][i];
		square += SQ (FM[run][i]);
	    }
	  mean /= nrrun;
	  square /= nrrun;
	  stdev = sqrt (square - SQ (mean));
	  printf
	      ("fraction of time machine %d produces at max rate is        %g   stdev %g\n",
	       i, mean, stdev);
	  for (mean = square = 0.0, run = 1; run <= nrrun; run++)
	    {
		mean += VM[run][i];
		square += SQ (VM[run][i]);
	    }
	  mean /= nrrun;
	  square /= nrrun;
	  stdev = sqrt (square - SQ (mean));
	  printf
	      ("mean duration of period machine %d produces at max rate is %g   stdev %g\n",
	       i, mean, stdev);
      }
}

void initialization ()
{
    int     i,
            j;

    srand48 (seed);

    t = 0.0;
    totalprod = 0.0;

    stmach[0] = UP;
    evmach[0] = t + uptime (0);
    v[0] = vmax[0];
    vmaxlength[0] = 0.0;
    vmaxnr[0] = 1;
    for (i = 1; i <= N; i++)
      {
	  stmach[i] = UP;
	  evmach[i] = t + uptime (i);
	  v[i] = MIN (vmax[i], q[i - 1] * v[i - 1]);
	  vmaxlength[i] = 0.0;
	  vmaxnr[i] = (v[i] < vmax[i]) ? 0 : 1;

	  b[i] = 0.0;
	  if (v[i] > 0)
	      stbuf[i] = NEUTRAL;
	  else
	      stbuf[i] = EMPTY;
	  area1[i] = 0.0;
	  area2[i] = 0.0;
	  vnet[i] = q[i - 1] * v[i - 1] - v[i];
	  if (vnet[i] > 0.0)
	      evbuf[i] = t + K[i] / vnet[i];
	  else
	      evbuf[i] = INF;
      }

    nrsample = 0;
    tsample = runlength / MAXS;
    for (i = 1; i <= N; i++)
	for (j = 0; j < MAXI; j++)
	    histo[i][j] = 0.0;
}

void updatestat (int run)
{						/* update statistics and reset */
    int     i;

    if (run > 0)
      {						/* ignore first run */
	  TH[run] = totalprod / t;
	  for (i = 1; i <= N; i++)
	    {
		L1[run][i] = area1[i] / t;
		L2[run][i] = area2[i] / t;
		LS[run][i] = sqrt (L2[run][i] - SQ (L1[run][i]));
	    }
	  for (i = 0; i <= N; i++)
	    {
		FM[run][i] = vmaxlength[i] / t;
		VM[run][i] = vmaxlength[i] / vmaxnr[i];
	    }
      }

    totalprod = 0.0;				/* reset */
    for (i = 1; i <= N; i++)
      {
	  area1[i] = 0.0;
	  area2[i] = 0.0;
      }
    for (i = 0; i <= N; i++)
      {
	  vmaxlength[i] = 0.0;
	  vmaxnr[i] = (v[i] < vmax[i]) ? 0 : 1;
      }

    for (i = 0; i <= N; i++)
	evmach[i] -= t;
    for (i = 1; i <= N; i++)
	evbuf[i] -= t;
    tsample -= t;
    t = 0.0;
}

int nextevent (int *j)
{
    int     i,
            jm,
            jb;

    jm = 0;
    for (i = 1; i <= N; i++)
	if (evmach[i] < evmach[jm])
	    jm = i;

    jb = 1;
    for (i = 1; i <= N; i++)
	if (evbuf[i] < evbuf[jb])
	    jb = i;

    if (evmach[jm] < evbuf[jb])
      {
	  *j = jm;
	  if (evmach[jm] < tsample)
	      return MACH;
	  else
	      return HIS;
      }
    else
      {
	  *j = jb;
	  if (evbuf[jb] < tsample)
	      return BUF;
	  else
	      return HIS;
      }
}

void update ()
{
    int     i;
    double  bprev;

    for (i = 1; i <= N; i++)
      {
	  bprev = b[i];
	  b[i] += vnet[i] * (t - tprev);	/* update buffer contents */
	  if (b[i] < -EPS || b[i] > K[i] + EPS)
	    {					/* check whether buffer levels are still between the boundaries */
		printf ("level of buffer %d is %g: infeasible", i, b[i]);
		exit (1);
	    }
	  area1[i] += (t - tprev) * (b[i] + bprev) / 2.0;	/* update area of buffer content process */
	  area2[i] +=
	      (t - tprev) * (SQ (b[i]) + b[i] * bprev + SQ (bprev)) / 3.0;
      }

    if (stmach[N] == UP)			/* update total production */
	totalprod += (t - tprev) * q[N] * v[N];
    for (i = 0; i <= N; i++)			/* update duration machine produces at max rate */
	if (v[i] >= vmax[i])
	    vmaxlength[i] += t - tprev;
}

void updatemach (int j)
{
    int     i;

    if (stmach[j] == UP)
      {						/* state changes from up to down */
	  stmach[j] = DOWN;
	  evmach[j] = t + downtime (j);
	  v[j] = 0.0;
	  if (j > 0)
	    {					/* update upstream line */
		for (i = j; i > 0 && stbuf[i] == FULL; i--)
		  {				/* buffer is full */
		      vnet[i] = 0.0;
		      evbuf[i] = INF;
		      v[i - 1] = 0.0;
		      evres[i - 1] = evmach[i - 1] - t;
		      if (stmach[i - 1] != UP)
			{			/* upstream machine SHOULD BE UP, otherwise buffer cannot be full */
			    printf
				("Machine %d is not up, while buffer %d is full and machine %d is UP: not possible\n",
				 i - 1, i, i);
			    exit (1);
			}
		      stmach[i - 1] = BLOCKED;	/* change status from UP to BLOCKED */
		      evmach[i - 1] = INF;
		  }
		if (i > 0)
		  {				/* buffer is not full */
		      vnet[i] = q[i - 1] * v[i - 1] - v[i];	/* v[i] == 0.0 */
		      if (vnet[i] > 0.0)
			{
			    evbuf[i] = t + (K[i] - b[i]) / vnet[i];
			    stbuf[i] = NEUTRAL;
			}
		      else			/* vnet[i] == 0.0 */
			  evbuf[i] = INF;
		  }
	    }
	  if (j < N)
	    {					/* update downstream line */
		for (i = j; i < N && stbuf[i + 1] == EMPTY; i++)
		  {				/* buffer is empty */
		      vnet[i + 1] = 0.0;
		      evbuf[i + 1] = INF;
		      if (stmach[i + 1] != UP)
			{			/* downstream machine SHOULD BE UP, otherwise buffer cannot be empty */
			    printf
				("Machine %d is not up, while buffer %d is empty and machine %d is UP: not possible\n",
				 i + 1, i + 1, i);
			    exit (1);
			}
		      v[i + 1] = 0.0;
		      evres[i + 1] = evmach[i + 1] - t;
		      stmach[i + 1] = STARVED;	/* change status from UP to STARVED */
		      evmach[i + 1] = INF;
		  }
		if (i < N)
		  {				/* buffer is not empty */
		      vnet[i + 1] = q[i] * v[i] - v[i + 1];	/* v[i] == 0.0 */
		      if (vnet[i + 1] < 0.0)
			{
			    evbuf[i + 1] = t - b[i + 1] / vnet[i + 1];
			    stbuf[i + 1] = NEUTRAL;
			}
		      else			/* vnet[i+1] == 0.0 */
			  evbuf[i + 1] = INF;
		  }
	    }
      }
    else if (stmach[j] == DOWN)
      {						/* state changes from down to up */
	  if (j > 0 && stbuf[j] == EMPTY)
	    {
		if (stmach[j - 1] == UP)
		  {				/* upstream machine cannot be UP, since the buffer is empty */
		      printf
			  ("Machine %d is up, while buffer %d is empty and machine %d is DOWN: not possible\n",
			   j - 1, j, j);
		      exit (1);
		  }
		stmach[j] = STARVED;
		evres[j] = uptime (j);
		evmach[j] = INF;
	    }
	  else if (j < N && stbuf[j + 1] == FULL)
	    {
		if (stmach[j + 1] == UP)
		  {				/* upstream machine cannot be UP, since the buffer is empty */
		      printf
			  ("Machine %d is up, while buffer %d is full and machine %d is DOWN: not possible\n",
			   j + 1, j + 1, j);
		      exit (1);
		  }
		stmach[j] = BLOCKED;
		evres[j] = uptime (j);
		evmach[j] = INF;
	    }
	  else
	    {
		stmach[j] = UP;
		evmach[j] = t + uptime (j);
		v[j] = vmax[j];
		vmaxnr[j]++;
		if (j > 0)
		  {				/* update upstream line */
		      for (i = j; i > 0 && stbuf[i] == FULL; i--)
			{			/* buffer is full and upstream machine is blocked */
			    if (stmach[i - 1] != BLOCKED)
			      {			/* upstream machine SHOULD BE BLOCKED, otherwise buffer cannot be full */
				  printf
				      ("Machine %d is not blocked, while buffer %d is full and machine %d is DOWN or BLOCKED: not possible\n",
				       i - 1, i, i);
				  exit (1);
			      }
			    v[i - 1] = MIN (vmax[i - 1], v[i]);
			    if (v[i - 1] >= vmax[i - 1])
				vmaxnr[i - 1]++;
			    stmach[i - 1] = UP;
			    evmach[i - 1] = t + evres[i - 1];
			    vnet[i] = q[i - 1] * v[i - 1] - v[i];
			    if (vnet[i] < 0.0)
			      {
				  evbuf[i] = t - b[i] / vnet[i];
				  stbuf[i] = NEUTRAL;
			      }
			    else		/* vnet[i] == 0.0 */
				evbuf[i] = INF;
			}
		      if (i > 0)
			{			/* buffer is not full nor empty */
			    if (stbuf[i] != NEUTRAL)
			      {
				  printf
				      ("Buffer %d is empty or full: not possible\n",
				       i);
				  exit (1);
			      }
			    vnet[i] = q[i - 1] * v[i - 1] - v[i];
			    if (vnet[i] > 0.0)
				evbuf[i] = t + (K[i] - b[i]) / vnet[i];
			    else if (vnet[i] < 0.0)
				evbuf[i] = t - b[i] / vnet[i];
			    else		/* vnet[i] == 0.0 */
				evbuf[i] = INF;
			}
		  }
		if (j < N)
		  {				/* update downstream line */
		      for (i = j; i < N && stbuf[i + 1] == EMPTY; i++)
			{			/* buffer is empty and downstream machine is starved */
			    if (stmach[i + 1] != STARVED)
			      {			/* downstream machine SHOULD BE STARVED, otherwise buffer cannot be empty */
				  printf
				      ("Machine %d is not starved, while buffer %d is empty and machine %d is DOWN or STARVED: not possible\n",
				       i + 1, i + 1, i);
				  exit (1);
			      }
			    v[i + 1] = MIN (q[i] * v[i], vmax[i + 1]);
			    if (v[i + 1] >= vmax[i + 1])
				vmaxnr[i + 1]++;
			    stmach[i + 1] = UP;
			    evmach[i + 1] = t + evres[i + 1];
			    vnet[i + 1] = q[i] * v[i] - v[i + 1];
			    if (vnet[i + 1] > 0.0)
			      {
				  evbuf[i + 1] = t + K[i + 1] / vnet[i + 1];
				  stbuf[i + 1] = NEUTRAL;
			      }
			    else		/* vnet[i+1] == 0.0 */
				evbuf[i + 1] = INF;
			}
		      if (i < N)
			{			/* buffer is not empty nor full */
			    if (stbuf[i + 1] != NEUTRAL)
			      {
				  printf
				      ("Buffer %d is empty or full: not possible\n",
				       i + 1);
				  exit (1);
			      }
			    vnet[i + 1] = q[i] * v[i] - v[i + 1];
			    if (vnet[i + 1] < 0.0)
				evbuf[i + 1] = t - b[i + 1] / vnet[i + 1];
			    else if (vnet[i + 1] > 0.0)
				evbuf[i + 1] =
				    t + (K[i + 1] - b[i + 1]) / vnet[i + 1];
			    else		/* vnet[i+1] == 0.0 */
				evbuf[i + 1] = INF;
			}
		  }
	    }

      }
    else
      {						/* machine[j] is nor UP or DOWN: not possible */
	  printf ("machine %d is nor UP or DOWN: not possible\n");
	  exit (1);
      }
}

void updatebuf (int j)
{
    int     i;

    if (vnet[j] < 0.0)
	stbuf[j] = EMPTY;
    else if (vnet[j] > 0.0)
	stbuf[j] = FULL;
    else
      {
	  printf ("Event in buffer %d while vnet[%d] = 0: impossible", j, j);
	  exit (1);
      }

    vnet[j] = 0.0;
    evbuf[j] = INF;

    if (stbuf[j] == EMPTY)
      {						/* machine j and (possibly) downstream machines have to slow down */
	  for (i = j; i <= N && stbuf[i] == EMPTY; i++)
	    {					/* buffer becomes empty */
		if (stmach[i] != UP)
		  {				/* downstream machine SHOULD BE UP, otherwise buffer cannot become empty */
		      printf
			  ("Machine %d is not up, while buffer %d becomes empty: not possible\n",
			   i, i);
		      exit (1);
		  }
		v[i] = q[i - 1] * v[i - 1];	/* machine i slows down */

		if (v[i] <= 0.0)
		  {
		      evres[i] = evmach[i] - t;
		      stmach[i] = STARVED;
		      evmach[i] = INF;
		  }
	    }
	  if (i <= N)
	    {					/* buffer i is not emtpy */
		if (stbuf[i] == FULL)
		    if (q[i - 1] * v[i - 1] - v[i] >= 0.0)
			vnet[i] = 0.0;
		    else
		      {
			  vnet[i] = q[i - 1] * v[i - 1] - v[i];
			  evbuf[i] = t - b[i] / vnet[i];
			  stbuf[i] = NEUTRAL;
		      }
		else
		  {
		      vnet[i] = q[i - 1] * v[i - 1] - v[i];
		      if (vnet[i] < 0.0)
			  evbuf[i] = t - b[i] / vnet[i];
		      else if (vnet[i] > 0.0)
			  evbuf[i] = t + (K[i] - b[i]) / vnet[i];
		      else			/* vnet[i] == 0.0 */
			  evbuf[i] = INF;
		  }
	    }
      }
    else if (stbuf[j] == FULL)
      {						/* machine j-1 and (possibly) upstream machines have to slow down */

	  for (i = j; i > 0 && stbuf[i] == FULL; i--)
	    {					/* buffer becomes full */
		if (stmach[i - 1] != UP)
		  {				/* upstream machine SHOULD BE UP, otherwise buffer cannot become full */
		      printf
			  ("Machine %d is not up, while buffer %d becomes full: not possible\n",
			   i - 1, i);
		      exit (1);
		  }
		v[i - 1] = v[i] / q[i - 1];	/* machine i-1 slows down */
		if (v[i - 1] <= 0.0)
		  {
		      evres[i - 1] = evmach[i - 1] - t;
		      stmach[i - 1] = BLOCKED;
		      evmach[i - 1] = INF;
		  }
	    }
	  if (i > 0)
	    {					/* buffer i is not full */
		if (stbuf[i] == EMPTY)
		    if (q[i - 1] * v[i - 1] - v[i] <= 0.0)
			vnet[i] = 0.0;
		    else
		      {
			  vnet[i] = q[i - 1] * v[i - 1] - v[i];
			  evbuf[i] = t + (K[i] - b[i]) / vnet[i];
			  stbuf[i] = NEUTRAL;
		      }
		else
		  {
		      vnet[i] = q[i - 1] * v[i - 1] - v[i];
		      if (vnet[i] < 0.0)
			  evbuf[i] = t - b[i] / vnet[i];
		      else if (vnet[i] > 0.0)
			  evbuf[i] = t + (K[i] - b[i]) / vnet[i];
		      else			/* vnet[i] == 0.0 */
			  evbuf[i] = INF;

		  }
	    }
      }
}

double uptime (int i)
{
    int     k,
            l;
    double  c2,
            p,
            p1,
            p2,
            mu,
            mu1,
            mu2,
            m1,
            m2,
            m3,
            a1,
            a2,
            x;

    c2 = SQ (upsd[i] / upmean[i]);

    if (c2 <= 1.0)
      {						/* fit mixture of Erlang-(k-1) and Erlang-k */
	  for (k = 2; (1.0 / k) > c2; k++);
	  p = (k * c2 - sqrt (k * (1.0 + c2) - SQ (k) * c2)) / (1.0 + c2);
	  mu = (k - p) / upmean[i];
	  for (x = 0.0, l = 1; l < k; x += exponential (mu), l++);
	  if (drand48 () > p)
	      x += exponential (mu);
      }
    else
      {						/* fit Hyper-exponential on first three moments, if possible */
	  m1 = upmean[i];
	  m2 = SQ (upsd[i]) + SQ (upmean[i]);
	  m3 = upm3[i];

	  if (m1 * m3 >= 1.5 * SQ (m2))
	    {
		a2 = (6.0 * SQ (m1) - 3.0 * m2) / (1.5 * SQ (m2) - m1 * m3);
		a1 = (1.0 + 0.5 * m2 * a2) / m1;
		mu1 = 0.5 * (a1 + sqrt (SQ (a1) - 4.0 * a2));
		mu2 = 0.5 * (a1 - sqrt (SQ (a1) - 4.0 * a2));
		p1 = mu1 * (1.0 - mu2 * m1) / (mu1 - mu2);
		p2 = 1.0 - p1;
	    }
	  else
	    {					/* fit Hyper-exponential with balanced means */
		p1 = (1.0 + sqrt ((c2 - 1.0) / (c2 + 1.0))) / 2.0;
		p2 = 1.0 - p1;
		mu1 = 2.0 * p1 / upmean[i];
		mu2 = 2.0 * p2 / upmean[i];
	    }
	  if (drand48 () < p1)
	      x = exponential (mu1);
	  else
	      x = exponential (mu2);
      }
    return x;
}

double downtime (int i)
{
    int     k,
            l;
    double  c2,
            p,
            p1,
            p2,
            mu,
            mu1,
            mu2,
            m1,
            m2,
            m3,
            a1,
            a2,
            x;

    c2 = SQ (downsd[i] / downmean[i]);

    if (c2 <= 1.0)
      {						/* fit mixture of Erlang-(k-1) and Erlang-k */
	  for (k = 2; (1.0 / k) > c2; k++);
	  p = (k * c2 - sqrt (k * (1.0 + c2) - SQ (k) * c2)) / (1.0 + c2);
	  mu = (k - p) / downmean[i];
	  for (x = 0.0, l = 1; l < k; x += exponential (mu), l++);
	  if (drand48 () > p)
	      x += exponential (mu);
      }
    else
      {						/* fit Hyper-exponential on first three moments, if possible */
	  m1 = downmean[i];
	  m2 = SQ (downsd[i]) + SQ (downmean[i]);
	  m3 = downm3[i];

	  if (m1 * m3 >= 1.5 * SQ (m2))
	    {
		a2 = (6.0 * SQ (m1) - 3.0 * m2) / (1.5 * SQ (m2) - m1 * m3);
		a1 = (1.0 + 0.5 * m2 * a2) / m1;
		mu1 = 0.5 * (a1 + sqrt (SQ (a1) - 4.0 * a2));
		mu2 = 0.5 * (a1 - sqrt (SQ (a1) - 4.0 * a2));
		p1 = mu1 * (1.0 - mu2 * m1) / (mu1 - mu2);
		p2 = 1.0 - p1;
	    }
	  else
	    {					/* fit Hyper-exponential with balanced means */
		p1 = (1.0 + sqrt ((c2 - 1.0) / (c2 + 1.0))) / 2.0;
		p2 = 1.0 - p1;
		mu1 = 2.0 * p1 / downmean[i];
		mu2 = 2.0 * p2 / downmean[i];
	    }
	  if (drand48 () < p1)
	      x = exponential (mu1);
	  else
	      x = exponential (mu2);
      }
    return x;
}

double exponential (double mu)
{						/* sample from exponential
						 * distribution with parameter mu */
    return -log (drand48 ()) / mu;
}

void printstatus ()
{
    int     k;

    printf ("current time is %g\n", t);
    printf ("status machines:\n");
    for (k = 0; k <= N; k++)
	printf ("event in machine %d at %g\n", k, evmach[k]);
    for (k = 0; k <= N; k++)
	printf ("status machine %d is %d\n", k, stmach[k]);
    for (k = 0; k <= N; k++)
	printf ("speed machine %d is %g\n", k, v[k]);
    printf ("status buffers:\n");
    for (k = 1; k <= N; k++)
	printf ("event in buffer %d at %g\n", k, evbuf[k]);
    for (k = 1; k <= N; k++)
	printf ("level in buffer %d is %g\n", k, b[k]);
    for (k = 1; k <= N; k++)
	printf ("net speed buffer %d is %g\n", k, vnet[k]);
}

void updatehisto ()
{
    int     i,
            j;

    nrsample++;
    tsample += runlength / MAXS;

    for (i = 1; i <= N; i++)
      {
	  for (j = 1; j < MAXI && b[i] >= j * K[i] / MAXI; j++);
	  histo[i][j - 1]++;
      }
}

void printhisto ()
{
    int     i,
            j;

    for (i = 1; i <= N; i++)
      {
	  printf ("buffer %d content distribution\n", i);
	  for (j = 0; j < MAXI; j++)
	      printf ("%g, %g\n", j * K[i] / MAXI,
		      (double) histo[i][j] / nrsample);
      }

}
